// Mr Watch DCC Monitor
// By yaasan
// Copyright(C) 2017- Desktop Station Co.,Ltd.

#include "NmraDcc_v141.h"
#include <Wire.h>
#include "I2CLCDLib.h"

//各種設定、宣言

//#define DEBUG

#define PIN_DCCIN 2

#define DECODER_ADDRESS 3
#define MAX_LOCTABLE 16
#define MAX_ACCTABLE 128

#define MODE_LOC 0
#define MODE_ACC 1
#define MODE_VMON 2
#define MODE_ERR 3
#define MODE_TOTAL 4

#define PULSE_MIN	40
#define PULSE_MAX	200


//使用クラスの宣言
NmraDcc	 Dcc;
DCC_MSG	 Packet;

/* LCD functions */
I2CLCDLib LCD;

//Task Schedule
unsigned long gPreviousL5 = 0;
unsigned long gPreviousL1 = 0;


struct LocData{
  uint16_t	mAddr;
  uint16_t	mSpeed;
  uint8_t	mDirection;
  uint32_t	mFunction;
};



LocData gLocTable[MAX_LOCTABLE];
uint16_t gAccTable[MAX_ACCTABLE];

uint8_t gMode = MODE_LOC;
uint8_t gTopIndex_Loc = 0xFF;
uint8_t gTopIndex_Acc = 0;

uint8_t gWatchMode = 0; // Analogを初期モードにする

uint8_t gButtonStatus = 0;

uint16_t gCount_Err = 0;
uint16_t gCount_Idle = 0;
uint16_t gCount_Reset = 0;
uint16_t gCount_PreMin = 255;
uint16_t gCount_PreMax = 0;


void(* resetFunc) (void) = 0;  //declare reset function at address 0
void OutputResults();
void ClearTable();
void DisplayFunctionBit(uint32_t inFunctionBits);
void DisplayAccBit(uint16_t inFunctionBits);
uint16_t ReadEdcVoltage();

void updateDccSpeed(word inLocoAddr, uint16_t inAddr, uint8_t inSpeed, uint8_t inDir, uint8_t inSpeedSteps );
void decodeLocoPacket(uint8_t inPacket1, uint8_t inPacket2);
uint16_t getLocAddrIndex(uint16_t inAddr);


void setup()
{
	ClearTable();
	
	//Pin mode
	pinMode(A2, INPUT); //Edc
	pinMode(A3, INPUT); //Button
	digitalWrite(A3, 1); //pullup
	
	
	pinMode(A0, OUTPUT);
	digitalWrite(A0, 0); //Reset High
	delay(50);
	digitalWrite(A0, 1); //Reset High
	delay(50);
	
	
	/* LCD init */
	LCD.begin();

	LCD.lcd_setCursor(0, 0);
	LCD.lcd_printStr("Desktop ");
	LCD.lcd_setCursor(0, 1);
	LCD.lcd_printStr(" Station"); 
	
	
	//シリアル通信開始
	Serial.begin(115200);

	Serial.println("--------------------------------------");
	Serial.println("Desktop Station Monitor               ");
	Serial.println("--------------------------------------");
	Serial.println("100 Ready");  
	Serial.println("@DSG,999,");

	// Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
	Dcc.pin(0, PIN_DCCIN, 0);

	// Call the main DCC Init function to enable the DCC Receiver
	Dcc.init( MAN_ID_DIY, 10, 0 , 0 ); 

	//Reset task
	gPreviousL5 = millis();

}

void loop()
{
	uint8_t aUpdate = 0;
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	
	if( (millis() - gPreviousL1) >= 100)
	{
		
		uint8_t aButton = digitalRead(A3);
		
		//Update Button status
		if( (gButtonStatus == 1) && (aButton == 0))
		{
			gMode++;
			aUpdate = 1;
			
		}
		
		gButtonStatus = aButton;
		
		//Reset task
		gPreviousL1 = millis();
		
	}
	
	if( ((millis() - gPreviousL5) >= 1000) || (aUpdate == 1))
	{
		//Main Task
		OutputResults();
		
		//Reset task
		gPreviousL5 = millis();
		
	}
}

/* Main Task (2sec) */
void OutputResults()
{
	int aCount = 0;
	char aTexts[9];
	uint16_t aVolt;
	uint16_t aLocAddr;
	uint16_t aMax0 = 0;
	uint16_t aMax1 = 0;
	uint16_t aMin0 = 0xFFFF;
	uint16_t aMin1 = 0xFFFF;
	uint16_t aPulse;
	
	if( gWatchMode == 0 )
	{
		if( gMode > 2)
		{
			gMode = 0;
		}
		
		switch( gMode)
		{
			
		case 0:
			LCD.lcd_setCursor(0, 0);
			LCD.lcd_printStr("NO  DCC ");
			LCD.lcd_setCursor(0, 1);
			LCD.lcd_printStr("SIGNALS ");
			break;
			
		case 1:
			aVolt = ReadEdcVoltage();
			sprintf(aTexts, "%2d.%dV   ", aVolt / 10, aVolt % 10);
			LCD.lcd_setCursor(0, 0);
			LCD.lcd_printStr(aTexts);
			LCD.lcd_setCursor(0, 1);
			LCD.lcd_printStr("-ANALOG-");
			break;
			
		case 2:
			gMode = 0;
			break;
			
		}
		
	}
	else
	{
		
		if( gMode > 4)
		{
			gMode = 0;
		}
		
		
		switch( gMode)
		{
			
		case 0:
			
			if( gTopIndex_Loc != 0xFF)
			{
				/* ロコが登録中の時 */
				aLocAddr = (gLocTable[gTopIndex_Loc].mAddr);
				sprintf(aTexts, "L%4d%s%2d", aLocAddr, gLocTable[gTopIndex_Loc].mDirection == 0 ? "R" : "F", gLocTable[gTopIndex_Loc].mSpeed);
				LCD.lcd_setCursor(0, 0);
				LCD.lcd_printStr(aTexts);
				
				LCD.lcd_setCursor(0, 1);
				for( int i = 0; i < 8; i++)
				{
					if( ((gLocTable[gTopIndex_Loc].mFunction >> i ) & 1) == 0)
					{
						aTexts[i] = '_';
					}
					else
					{
						aTexts[i] = 'O';
					}
				}
				LCD.lcd_printStr(aTexts);
			}
			else
			{
				/* ロコのパケットが流れてないとき */
				sprintf(aTexts, "Idle %3d", gCount_Idle);
				LCD.lcd_setCursor(0, 0);
				LCD.lcd_printStr(aTexts);
				LCD.lcd_setCursor(0, 1);
				LCD.lcd_printStr("  -DCC- ");
				
				/* アイドルパケットをリセット */
				gCount_Idle = 0;
			}
			
			break;
			
		case 1:
			//Accessory
			for( int i = 0; i < 16; i++)
			{
				LCD.lcd_setCursor(i & 0b111, i >> 3);
				if( ((gAccTable[0] >> i ) & 1) == 0)
				{
					LCD.lcd_printStr("/");
				}
				else
				{
					LCD.lcd_printStr("|");
				}
			}
			break;
			
		case 2:
			//Edc
			aVolt = ReadEdcVoltage();
			sprintf(aTexts, "%2d.%dV   ", aVolt / 10, aVolt % 10);
			LCD.lcd_setCursor(0, 0);
			LCD.lcd_printStr(aTexts);
			LCD.lcd_setCursor(0, 1);
			LCD.lcd_printStr("  -DCC- ");
			break;
			
		case 3:
			
			aMax0 = 0;
			aMax1 = 0;
			aMin0 = 0xFFFF;
			aMin1 = 0xFFFF;
			
			for( int i = 0; i < 100; i++)
			{
				aPulse = pulseIn(PIN_DCCIN, LOW, 200);
				
				if( (aPulse != 0) && (aPulse > 80) && (aPulse <= 200))
				{
					
					if( aPulse > aMax0)
					{
						aMax0 = aPulse;
					}
					else if ( aPulse < aMin0)
					{
						aMin0 = aPulse;
					}
				}
			}
			
			for( int i = 0; i < 100; i++)
			{
				aPulse = pulseIn(PIN_DCCIN, LOW, 100);
				
				if( (aPulse != 0) && (aPulse >= 48) && (aPulse <= 68))
				{
					if( aPulse > aMax1)
					{
						aMax1 = aPulse;
					}
					else if ( aPulse < aMin1)
					{
						aMin1 = aPulse;
					}
				}
			}
			
			sprintf(aTexts, "|%3d/%2d", aMin1, aMax1);
			LCD.lcd_setCursor(0, 0);
			LCD.lcd_printStr(aTexts);
			
			sprintf(aTexts, "O%3d/%2d", aMin0, aMax0);
			LCD.lcd_setCursor(0, 1);
			LCD.lcd_printStr(aTexts);
			break;
			
		case 4:
			
			sprintf(aTexts, "Pre%2d/%2d", gCount_PreMin, gCount_PreMax);
			LCD.lcd_setCursor(0, 0);
			LCD.lcd_printStr(aTexts);
			
			sprintf(aTexts, "Err%5d", gCount_Err);
			LCD.lcd_setCursor(0, 1);
			LCD.lcd_printStr(aTexts);
			
			break;
			
			
			
		}
	}
}

void ClearTable()
{
	for( int i = 0; i < MAX_LOCTABLE; i++)
	{
			gLocTable[i].mAddr = 0;
			gLocTable[i].mSpeed = 0;
			gLocTable[i].mDirection = 0;
			gLocTable[i].mFunction = 0;
	}
}

void DisplayFunctionBit(uint32_t inFunctionBits)
{
	
	for( int i = 0; i < 29; i++)
	{
		if( bitRead(inFunctionBits, i) == 0)
		{
				Serial.print(".");
		}
		else
		{
				Serial.print("O");
		}
		
	}
}

void DisplayAccBit(uint16_t inFunctionBits)
{
	
	for( int i = 0; i < 16; i++)
	{
		if( bitRead(inFunctionBits, i) == 0)
		{
				Serial.print(".");
		}
		else
		{
				Serial.print("O");
		}
		
	}
}

/* 全DCCメッセージを受信 */
void notifyDccMsg( DCC_MSG * Msg)
{
	uint8_t aErrorbyte = 0;
	uint16_t aLocoAddr = 0;
	uint16_t aAccAddr = 0;
	
	
	
	/* エラーチェック */
	for(uint8_t i = 0; i < Msg->Size; i++)
	{
		aErrorbyte = aErrorbyte ^ Msg->Data[i];
	}
	
	if( aErrorbyte != 0)
	{
		if( gCount_Err <= 65535)
		{
			gCount_Err++;
		}
		
		return;
	}
	else
	{
		gWatchMode = 1;
	}
	
	/* プリアンブル数計算 */
	if( Msg->PreambleBits < gCount_PreMin)
	{
		gCount_PreMin = Msg->PreambleBits;
	}
	else if( Msg->PreambleBits > gCount_PreMax)
	{
		gCount_PreMax = Msg->PreambleBits;
	}
	
		
	/* パケットチェック */
	switch(Msg->Data[0])
	{
	case 0x00:
		//Broadcast
		
		if( Msg->Data[1] == 0x00)
		{
			gCount_Reset++;
		}
		else
		{
			/* その他のBroadcast処理 */
		}
		
		break;
		
	case 0xFF:
		//Idle packet
		gCount_Idle++;
		break;
		
	default:
		
		if( Msg->Data[0] <= 127)
		{
			/* SHORTアドレスロコの操作 */
			aLocoAddr = Msg->Data[0];
			
			decodeLocoPacket(aLocoAddr, Msg->Data[1], Msg->Data[2]);
			
			
		}
		else if(( Msg->Data[0] >= 128) && ( Msg->Data[0] <= 191))
		{
			/* アクセサリ */
			aAccAddr = (((Msg->Data[0] & 0b111111) - 1) * 4) + ((7 - (Msg->Data[1] >> 4) & 0b111) << 8) +  (((Msg->Data[1] >> 1) & 0b11));
			
			if( (Msg->Data[1] & 1) == 0)
			{
				gAccTable[ aAccAddr >> 4 ] = gAccTable[ aAccAddr >> 4 ] & (~(1 << (aAccAddr & 0xFFFF)) & 0xFFFF);
				//uiPacketInfo.innerText = uiPacketInfo.innerText + ", direction is diverse.";
			}
			else
			{
				gAccTable[ aAccAddr >> 4 ] = gAccTable[ aAccAddr >> 4 ] | (1 << (aAccAddr & 0xFFFF)) ;
				//uiPacketInfo.innerText = uiPacketInfo.innerText + ", direction is straight.";
			}
		}
		else if(( Msg->Data[0] >= 192) && ( Msg->Data[0] <= 231))
		{
			/* LONGアドレスロコの操作 */
			aLocoAddr = ((Msg->Data[0] - 192) << 8) + Msg->Data[1];
			
			decodeLocoPacket(aLocoAddr, Msg->Data[2], Msg->Data[3]);
		}
		else if(( Msg->Data[0] >= 232) && ( Msg->Data[0] <= 254))
		{
			/* 予約 */
			
		}
		
		
		
		break;
		
	}
	
}

void decodeLocoPacket(word inLocoAddr, uint8_t inPacket1, uint8_t inPacket2)
{
	uint8_t aSpeed;
	uint8_t aDir;
	uint8_t aIndex = getLocAddrIndex(inLocoAddr);
	
	switch(inPacket1 >> 5)
	{
	case 0b000:
		break;
	
	case 0b001:
		
		switch(inPacket1 & 0b11111)
		{
			case 31:
				
				aSpeed = inPacket2 & 0b01111111;
				
				if( aSpeed > 1)
				{
					aSpeed = aSpeed - 1;
				}
				
			/* リスト更新 */
				updateDccSpeed( inLocoAddr, aSpeed, (inPacket2 >> 7) & 1, 127 );
			break;
		}
		break;

	case 0b010:
		//REV
			
		aSpeed = inPacket1 * 0b11111;
		
		if( aSpeed > 1)
		{
			aSpeed = aSpeed - 1;
		}
		
		/* リスト更新 */
		updateDccSpeed( inLocoAddr, aSpeed, 0, 27 );
	
		break;
	case 0b011:
		//FWD
			
		aSpeed = inPacket1 * 0b11111;
		
		if( aSpeed > 1)
		{
			aSpeed = aSpeed - 1;
		}
		
		/* リスト更新 */
		updateDccSpeed( inLocoAddr, aSpeed, 1, 27 );
		
		break;
		
	case 0b100:
		bitWrite( gLocTable[aIndex].mFunction, 0, (inPacket1 >> 4) & 1);
		bitWrite( gLocTable[aIndex].mFunction, 1, (inPacket1 >> 0) & 1);
		bitWrite( gLocTable[aIndex].mFunction, 2, (inPacket1 >> 1) & 1);
		bitWrite( gLocTable[aIndex].mFunction, 3, (inPacket1 >> 2) & 1);
		bitWrite( gLocTable[aIndex].mFunction, 4, (inPacket1 >> 3) & 1);
		break;
		
	case 0b101:
			
			switch((inPacket1 >> 4) & 1)
			{
			case 1:
				bitWrite( gLocTable[aIndex].mFunction, 5, (inPacket1 >> 0) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 6, (inPacket1 >> 1) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 7, (inPacket1 >> 2) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 8, (inPacket1 >> 3) & 1);
				break;
			case 0:
				bitWrite( gLocTable[aIndex].mFunction, 9, (inPacket1 >> 0) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 10, (inPacket1 >> 1) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 11, (inPacket1 >> 2) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 12, (inPacket1 >> 3) & 1);
				break;
			}
		break;
		
	case 0b110:
		
		
		switch(inPacket1 & 0b11111)
		{
		case 30:
				bitWrite( gLocTable[aIndex].mFunction, 13, (inPacket2 >> 0) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 14, (inPacket2 >> 1) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 15, (inPacket2 >> 2) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 16, (inPacket2 >> 3) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 17, (inPacket2 >> 4) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 18, (inPacket2 >> 5) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 19, (inPacket2 >> 6) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 20, (inPacket2 >> 7) & 1);
			break;
			
		case 31:
				bitWrite( gLocTable[aIndex].mFunction, 21, (inPacket2 >> 0) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 22, (inPacket2 >> 1) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 23, (inPacket2 >> 2) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 24, (inPacket2 >> 3) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 25, (inPacket2 >> 4) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 26, (inPacket2 >> 5) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 27, (inPacket2 >> 6) & 1);
				bitWrite( gLocTable[aIndex].mFunction, 28, (inPacket2 >> 7) & 1);
			break;
		}
	}
	

	
}

uint16_t getLocAddrIndex(uint16_t inAddr)
{
	uint16_t aIndex = 0xFFFF;
	
	for( int i = 0; i < MAX_LOCTABLE; i++)
	{
		if( (gLocTable[i].mAddr == inAddr) || (gLocTable[i].mAddr == 0) )
		{
			
			gLocTable[i].mAddr = inAddr;
			
			aIndex = i;
			
			break;
		}
	}
	
	return aIndex;
}

void updateDccSpeed( uint16_t inAddr, uint8_t inSpeed, uint8_t inDir, uint8_t inSpeedSteps )
{
	
	uint16_t aSpeed = 0;
	
	for( int i = 0; i < MAX_LOCTABLE; i++)
	{
		if( (gLocTable[i].mAddr == inAddr) || (gLocTable[i].mAddr == 0) )
		{
			
			gLocTable[i].mAddr = inAddr;
			
			if( inSpeed >= 1)
			{
				aSpeed = ((uint16_t)inSpeed * 100) / (uint16_t)inSpeedSteps;
			}
			else
			{
				aSpeed = 0;
			}
			
			
			if( (aSpeed != gLocTable[i].mSpeed) || (gLocTable[i].mDirection != inDir))
			{
				gLocTable[i].mSpeed = aSpeed;
				gLocTable[i].mDirection = inDir;
			
				gTopIndex_Loc = i;
			
			}
			
			/* for文終了 */
			break;
		}
		
		
	}
}


uint16_t ReadEdcVoltage()
{
	
	//電圧換算
	// 1k / 11k  y=((x*11*5)/1/1024)
	uint16_t aVolt = analogRead(A2);

	//Limit
	if( aVolt > 922)
	{
		//30Vで打ち止め
		aVolt = 922;
	}
	
	aVolt = aVolt >> 2;
	
	aVolt = (aVolt * 301) >> 7;
	
	if( aVolt > 255)
	{
		//UCに入るサイズにリミット(25.5V)
		aVolt = 255;
	}
	
	return aVolt;
	
}



